#-*-coding:utf-8-*-

"""fournit la classe CanalMono pour disposer d'un
collecteur de sons en RAM.
"""


import numpy as NY
from samplerate import sr, plus_petite_valeur
from conversions08 import extremum
    
try:
    from serie import Serie002
except ImportError:
    print "Le module 'serie' manque"
    print """Il faut excommenter les lignes qui appellent ce module;
          pratiquement toutes les lignes comprenant le mot 'alea'
          """

from origine import Origine
 
__author__ = """Copyright (c) 2005 René Bastian (Wissembourg, FRANCE)"""
__date__ = "2006.02.23, 07.09.15"



def filtreContinu(signal):
    """
    éliminer la composante continue du signal.
    """
    n = len(signal)
    w = NY.zeros(n)
    for i in range(1, n):
        w[i] = (signal[i] - signal[i-1]) * 0.5
    return w

class CanalMono(Origine):
    """
    permet de placer des vecteurs-sons en un point temporel précis ;
    cette classe est une des plus utilisées et quasiment indépendante
    de tout le reste de 'pythoneon', sauf :
    -- la fonction 'extremum' qui
       donne la plus grand valeur absolue dans un vecteur,
    -- la valeur 'sr' qui représente la fréquence d'échantillonnage,
    -- le contenu de 'numarray' qui pourrait être remplacé par 'numpy'.
    'fin' donne l'exacte mesure du signal - mais la différence n'est
    que de 1/10e de sec. Les grandes traînées quasi nulles proviennent
    des réverbs.
    """
    def __init__(self, dureeInitiale = 1.0, divbuffer=10,
                 couple = (1957, 9134)):
        """produit une instance d'une duree initiale donnée ;
        si cette durée initiale est insuffisante, elle est
        prolongée par paquets de 0.1 sec.
        La longueur du buffer est ajustable, soit
        par 'divbuffer' soit par la méthode 'setbuffer'."""
        Origine.__init__(self)
        self.buffer = NY.zeros(sr/divbuffer)
        n = int(dureeInitiale * sr)
        self.a = NY.zeros(n)
        self.len = len(self.a)
        self.alea = Serie002(*couple)
        self.fin = 0
        self.ncompte = 0

    def addT(self, pt, signal):
        """ajoute le 'signal' au point temporel 'pt'."""
        n =  int(pt*sr)
        ns = len(signal)
        
        dim = n + ns
        self.fin = max(self.fin, dim)
        if dim > self.len:
            k = dim - self.len + 1
            self.buffer = NY.zeros(k)
        while dim > self.len:
            self.a = NY.concatenate((self.a, self.buffer))
            self.len = len(self.a)
        try:
            self.a[n:n+ns] += signal
        except ValueError:
            print "CanalMono:addT:ValueError", "signal", len(signal),
            print "buffer", len(self.buffer), "n", n, "dim", dim
        else:
            self.ncompte += 1
    
        
    def addN(self, n, signal):
        """ajoute le signal à partir du sample d'indice 'n'."""
        ns = len(signal)
        dim = n + ns
        self.fin = max(self.fin, dim)
        if dim > self.len:
            k = dim - self.len + 1
            self.buffer = NY.zeros(k)
        while dim > self.len:
            self.a = NY.concatenate((self.a, self.buffer))
            self.len = len(self.a)
        self.a[n:n+ns] += signal
        self.ncompte += 1
        
    def ex(self, limite_continu=0.1):
        """
        retourne le vecteur-signal, normé sur 1.
        ATTENTION: ne doit être utilisé que sur des canaux uniques
        """
        w = self.a[:self.fin]/extremum(self.a[:self.fin])
        
        
        if abs(w.sum()/len(w)) > limite_continu: 
            return filtreContinu(w)
        else:
            return w


    def brut(self):
        """retourne le vecteur-signal, non normé"""
        return self.a[:self.fin]
    
    def dim(self):
        """retourne le nombre d'éléments du vecteur : len(self.a)"""        
        return self.fin

    def duree(self):
        """retourne la durée du vecteur 'self.a'"""
        return float(self.fin) / sr
    
    def extrait(self, n0, n1):
        """
        retourne la partie de vecteur allant de 'n0' à 'n1'
        en nombre de samples"""
        try:
            return self.a[n0:n1]
        except:
            print "Indice n1 incorrect : essai n0:"
            try :
                return self.a[n0:]
            except:
                print "Indice n0 incorrect !"
                return None

    def extraitT(self, t0, t1):
        """
        retourne la partie de vecteur allant de 't0' à 't1' en secondes
        """
        n0 = int(sr * t0)
        n1 = int(sr * t1)
        return self.extrait(n0, n1)
    
    def affiche(self):
        """affiche longueur et durée du vecteur."""
        print "Longueur %15i Duréee %10.5f" % (self.len, float(self.len)/sr)

    def tofile(self, nom):
        """pour écrire un fichier mf64."""
        self.a.tofile(nom + ".mf64")

    def close(self):
        """pour faire semblant de finir."""
        pass

    def controle_niveau(self):
        """
        détecter des plages de presque zéros à la fin d'un canal
        """
        i = 0
        empan = sr / 10
        dmax = extremum(self.a)/100 
        while i < self.len:
            dx = extremum(self.a[i:i+empan])
            if dx <= dmax:
                print "%12i %15.8f" % (i, dx)
            i += empan

    def autoreverb(self, premiercoeff = 0.125, duree = 2.0,
                   ecartmin = 0.04, ecartmax = 0.06,
                   pt = 0.0,
                   alea0 = 1955, alea1 = 9134):
        """
        Le contenu du canal, 'self.a', est additionné à
        lui-même en intercalant un délai et en appliquant
        un amortissement, ce qui simule une réverbération.        
        'pt'  : point temporel en sec où débute la réverbération        
        'premiercoeff' : float <=  1.0
        'duree' : durée de la réverb
        'ecartmin, ecartmax' : intervalle entre deux répétitions en sec

        utiliser:
        - premiercoeff = 0.625
        - duree = 3. ... 5.  
        """           
        epsilon = plus_petite_valeur
        alea = Serie002(alea0, alea1)
        ecartmoyen = (ecartmin + ecartmax) / 2.0
        coeff = NY.exp((NY.log(epsilon) - NY.log(premiercoeff)) * \
                    ecartmoyen / duree)
        nreflexions = int(NY.log(epsilon)/NY.log(coeff)) + 1
        Pts = alea.uniform(ecartmin, ecartmax, nreflexions)
        Pts = NY.add.accumulate(Pts) + pt
        signal = self.a
        signal = signal * premiercoeff                
        for ptr in Pts :
            self.addT(ptr, signal)
            signal = signal * coeff



class Reverberation:    
    """
    encore un essai de reverb de plus,
    mais on a plaisir à l'utiliser ;
    essayer d'imposer une enveloppe spéciale avec une attaque molle
    pour éviter les répétitions percussives
    (voir essai dans mazus:floeter)
    intégrer cette manoeuvre dans cette classe ??
    """ 
    def __init__(self, dureeInitiale = 1.0, 
                 premierecart = 0.050,                 
                 premiercoeff = 0.125,
                 duree = 4.,
                 ecartmin = 0.04,
                 ecartmax = 0.06,
                 epsilon = plus_petite_valeur,
                 couple=(96785,56347)):
        """
        installe une réverb ; les données initiales sont :
          premierecart = 0.050
          premiercoeff = 0.125
          duree = 4.    
          ecartmin = 0.04
          ecartmax = 0.06
        qu'on peut changer par 'change'.
        Il ne faut pas procéder par changement direct, car
        il faut faire effectuer les calculs par 'calcul'.
        """
        self.dureeInitiale = dureeInitiale
        self.premierecart = premierecart
        self.premiercoeff = premiercoeff
        self.duree = duree 
        self.ecartmin = ecartmin
        self.ecartmax = ecartmax 
        self.epsilon = epsilon
        self.alea = Serie002(*couple)
        ecartmoyen = (self.ecartmin + self.ecartmax) / 2.0
        self.coeff = NY.exp((NY.log(self.epsilon) - \
                             NY.log(self.premiercoeff)) * \
                            ecartmoyen / self.duree)
        self.nreflexions = int(NY.log(self.epsilon)/NY.log(self.coeff)) + 1

    def calcul(self):
        """
        effectue les calculs nécessaires après un changement
        des valeurs.
        """
        ecartmoyen = (self.ecartmin + self.ecartmax) / 2.0
        self.coeff = NY.exp((NY.log(self.epsilon) - \
                             NY.log(self.premiercoeff)) * \
                            ecartmoyen / self.duree)
        self.nreflexions = int(NY.log(self.epsilon)/NY.log(self.coeff)) + 1

    def __call__(self, signal):
        """
        retourne le signal réverbéré
        'signal' : le signal à réverbérer
        """
        self.c = CanalMono(self.dureeInitiale)
        signal *= self.premiercoeff
        self.c.addT(0.0, signal)
        Pts = self.alea.uniform(self.ecartmin, self.ecartmax, self.nreflexions)
        Pts = NY.add.accumulate(Pts)
        for pt in Pts:
            self.c.addT(pt, signal)
            signal = signal * self.coeff
        return self.c.a

    def filtre(self, signal, filtre):
        """
        retourne le signal réverbéré
        'signal' : le signal à réverbérer
        """
        self.c = CanalMono(self.dureeInitiale)
        signal *= self.premiercoeff
        self.c.addT(0.0, signal)
        Pts = self.alea.uniform(self.ecartmin, self.ecartmax, self.nreflexions)
        Pts = NY.add.accumulate(Pts)
        for pt in Pts:
            signal = filtre.filtrer(signal)
            self.c.addT(pt, signal)
            signal = signal * self.coeff
        return self.c.a


