#-*-coding:utf-8-*-

"""fournit la classe Partition qui a les avantages
de la partition traditionnelle : flexibilité du temps.

la version v0 comprend une gestion de hauteurs - mais c'est une
mauvaise idée : la partition doit placer des événements dans le temps,
et rien d'autre.

La classe Partition devrait aussi servir au montage:

On produit d'abord des fichiers 'float64'
- soit à l'aide de 'canal08.CanalMono().tofile(nom)'
- soit à l'aide de 'partition.Partition.close(efface=False)'
Pour un montage, il faut disposer:
- de la liste des fichiers
- pour chaque fichier
--- du point temporel 
--- de l'amplitude
Pour faire le montage:
- rédiger la liste de tuples (nom, ecart|pt, amplitude)
- instancier une 'Partition'
- appeler 'Partition.montage(Liste)'
- appeler 'Partition.close(efface=False)

La version 'v0-partition.py' utilise 'numarray'.
"""

import os
import numpy as NY
import utiles as _u

from enregistrer import Multi


__author__ = "Copyright (c) 2005-07 René Bastian (Wissembourg, FRANCE)"
__date__ = "2006.05.08, 07.08.24"

__all__ = ["Partition"]

class Partition:
    """classe permettant des opérations comparables à celles
    qui sont possibles avec une partition de musique habituelle."""
    def __init__(self, nomsouche, npistes, dureeInitiale = 1.0, informer=None,
                 nouveau = True, texte = None, titre=None):                 
        """
        'u' : unité (tempo)
        'tempo' : tempo (None si tempo fixe, fonction si tempovariable)
        'tempovar' : True si tempovariable 
        't' : point temporel        
        'npistes' : nombre de canaux
        'nomsouche' : nom des fichiers à réaliser
        'nouveau' :
            si True, l'existant est effacé
            si False, les signaux sont ajoutés à l'existant
        'texte' : destiné à être un nom de fichier-texte
          pour recueillir des données.
        """
        self.u = 1.0
        self.tempo = None
        self.pt0 = 0.0
        if len(nomsouche) > 0:
            self.nomsouche = nomsouche            
        else :
            print "Donner un nom-souche valable"
        self.npistes = npistes
        self.converti = False        
        self.collecteur = Multi(nomsouche, npistes, dureeInitiale, nouveau)
        self.t = 0.0
        self.w = None
        self.informer = informer
        self.titre = titre
        try:
            self.texte = open(texte, "w")
        except TypeError: 
            self.texte = texte

    
    
    def unit(self, u):
        """attribue la valeur u à l'unité de temps self.u."""
        self.u = u
        
    def pt(self, t):
        """donne un nouveau point temporel absolu
        TESTER TESTER
        """
        if self.tempo != None:
            self.u = self.tempo(self.pt0 - self.t)
        self.t = t * self.u
        
    def dr(self, ecart):
        """fixe le prochain point temporel en additiont 'ecart'
        au point temporel actuel."""
        if self.tempo != None:
            self.u = self.tempo(self.pt0 - self.t)
        self.t += ecart * self.u

    def ecart(self, xecart):
        """comme 'dr'"""
        if self.tempo != None:
            self.u = self.tempo(self.pt0 - self.t)
        self.t += xecart * self.u


    def dimension(self):
        """retourne la durée actuelle de la partition"""
        nom = self.nomsouche + ".0f64"
        return os.stat(nom).st_size/8

    
    
    def fevt(self, *fsignals):
        """placement des signaux : 'fsignals' est une liste de fichiers
        f64 représentant un événement sonore ;
        chaque 'fsignal' correspond à un haut-parleur.
        N'a jamais été utilisé avec 'numarray'.
        Utiliser 'montage'
        """
        for i, couple in enumerate(fsignals):
            nom, ampli = couple
            try:
                w = NY.fromfile(nom, NY.float64)
                w *= ampli
                self.collecteur.ajouter(self.t, i, w)
            except IOError: 
                print "Partition.fevt", nom, "injoignable."
                return

    def calculExtremum(self, *signals):
        """retourne l'extremum des signaux"""
        mx = 0.0
        for sig in signals:
            mx = max(mx, _u.extremum(sig))
        return mx

    def evt(self, *signals):
        """placement des signaux : 'signals' est une liste de signaux,
        chaque 'signals' correspond à un haut-parleur.        
        """
        forme = "%s : u %8.3f t %8.3f ampli %8.3f"
        
        try:
            for i in range(len(signals)):
                self.collecteur.ajouter(self.t, i, signals[i])
        except IOError:
            print "Partition.evt: pas de signal"
            return
        if self.informer:
            mx = self.calculExtremum(*signals)
            print forme % (self.titre, self.u, self.t, mx)


    def kt(self):
        """
        affiche unité et point temporel actuels
        """
        print "u %8.3f t %8.3f" % (self.u, self.t)

    def ptx(self):
        """
        retourne la valeur du point temporel
        """
        return self.t

    

    def efface_f64(self):
        """
        efface les fichiers en format Float64.
        """
        commande = " ".join(["rm", self.nomsouche + ".?f64"])
        os.system(commande)
    
    def close(self, efface = False):
        """
        écrit le fichier raw & le fichier wav ;
        efface (ad lib.) les f64        
        """
        try:
            self.texte.close()
        except AttributeError:
            print "Partition.texte == None"
        self.collecteur.fusion()
        commande = " ".join(["sox -c2 -sw -r44100", self.nomsouche+".raw",
                      "w"+self.nomsouche+".wav"])
        os.system(commande)
        if efface :
            self.efface_f64()
            

    def closeraw(self, efface = False):
        """
        écrit le fichier raw et efface (ad lib.) les f64
        """
        self.collecteur.fusion()        
        if efface :
            self.efface_f64()

    def closewav(self, nomwav=None, efface = False):
        """
        écrit le fichier wav et efface (ad lib.) les f64
        """        
        self.collecteur.fusion()
        if nomwav == None:
            commande = " ".join(["sox -c2 -sw -r44100", self.nomsouche+".raw",
                                 self.nomsouche+".wav"])
        else:
            commande = " ".join(["sox -c2 -sw -r44100", self.nomsouche+".raw",
                                 nomwav+".wav"])
        os.system(commande)
        
        if efface :
            self.efface_f64()

    def efface_raw(self):
        """
        efface les fichiers 'raw'
        """
        commande = " ".join(["rm", self.nomsouche + ".raw"])
        os.system(commande)


